from __future__ import absolute_import

import ctypes
import os
import sys

# un-comment this section to fix Cython backtrace line-numbers in
# IPython/Jupyter. see https://bugs.python.org/issue32797#msg323167
# ---
# try:
#    from importlib.machinery import ExtensionFileLoader
# except ImportError:
#    pass  # Python 2
# else:
#    del ExtensionFileLoader.get_source
# ---

if os.name == "posix":
    if sys.platform == "darwin":
        lib_name = "libtiledb.dylib"
    else:
        lib_name = "libtiledb.so"
else:
    lib_name = "tiledb"

# On Windows and whl builds, we may have a shared library already linked, or
# adjacent to, the cython .pyd shared object. In this case, we can import directly
# from .libtiledb
try:
    import tiledb
    from .libtiledb import Ctx
except:
    try:
        lib_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "native")
        ctypes.CDLL(os.path.join(lib_dir, lib_name))
    except OSError as e:
        # Otherwise try loading by name only.
        ctypes.CDLL(lib_name)

from .libtiledb import (
    Array,
    Ctx,
    Config,
    Dim,
    Domain,
    Attr,
    ArraySchema,
    TileDBError,
    VFS,
    FileIO,
    FilterList,
    NoOpFilter,
    GzipFilter,
    ZstdFilter,
    LZ4Filter,
    Bzip2Filter,
    RleFilter,
    DoubleDeltaFilter,
    BitShuffleFilter,
    ByteShuffleFilter,
    BitWidthReductionFilter,
    PositiveDeltaFilter,
    ChecksumMD5Filter,
    ChecksumSHA256Filter,
    consolidate,
    default_ctx,
    group_create,
    object_type,
    ls,
    walk,
    remove,
    move,
    schema_like,
    stats_enable,
    stats_disable,
    stats_reset,
    stats_dump,
    vacuum,
)

from .array import DenseArray, SparseArray

from ._fragment import info as fragment_info

from .highlevel import open, save, from_numpy, empty_like, array_exists

# TODO restricted imports
from .dataframe_ import from_csv, from_pandas, from_dataframe, open_dataframe
from .parquet_ import from_parquet

from .version import version as __version__

# Note: we use a modified namespace packaging to allow continuity of existing TileDB-Py imports.
#       Therefore, 'tiledb/__init__.py' must *only* exist in this package.
#       Furthermore, in sub-packages, the `find_packages` helper will not work at the
#       root directory due to lack of 'tiledb/__init__.py'. Sub-package 'setup.py' scripts
#       must declare constituents accordingly, such as by running 'find_packages' on a sub-directory
#       and applying prefixes accordingly.
#   1) https://packaging.python.org/guides/packaging-namespace-packages/#native-namespace-packages
#   2) https://stackoverflow.com/a/53486554
#
# Note: 'pip -e' in particular will not work without this declaration:
__path__ = __import__("pkgutil").extend_path(__path__, __name__)



from __future__ import absolute_import

import sys
import array as cparray
from tiledb import libtiledb as libmetadata
from tiledb.libtiledb import ustring


class Metadata(object):
    def __init__(self, array):
        self.array = array

    def __setitem__(self, key, value):
        """
        Implementation of [key] <- val (dict item assignment)

        :param key: key to set
        :param value: corresponding value
        :return: None
        """

        libmetadata.put_metadata(self.array, key, value)

    def __getitem__(self, key):
        """
        Implementation of [key] -> val (dict item retrieval)
        :param key:
        :return:
        """
        if not (isinstance(key, str) or isinstance(key, unicode)):
            raise ValueError(
                "Unexpected key type '{}': expected str " "type".format(type(key))
            )

        # `get_metadata` expects unicode
        key = ustring(key)
        v = libmetadata.get_metadata(self.array, key)

        if v is None:
            raise TileDBError("Failed to unpack value for key: '{}'".format(key))

        return v

    def __contains__(self, key):
        """
        Returns True if 'key' is found in metadata store.
        Provides support for python 'in' syntax ('k in A.meta')

        :param key: Target key to check.
        :return:
        """

        try:
            self[key]
        except KeyError:
            return False

        return True

    def consolidate(self):
        """
        Consolidate array metadata. Array must be closed.

        :return:
        """

        # TODO: ensure that the array is not x-locked?

        libmetadata.consolidate_metadata(self.array)

    def __delitem__(self, key):
        """
        Remove key from metadata.

        **Example:**

        >>> # given A = tiledb.open(uri, ...)
        >>> del A.meta['key']

        :param key:
        :return:
        """

        libmetadata.del_metadata(self.array, key)

    def __len__(self):
        """
        :return: length of metadata store
        """

        return libmetadata.len_metadata(self.array)

    def keys(self):
        """
        Return metadata keys as list.

        :return: List of array metadata keys.
        """

        return libmetadata.load_metadata(self.array, unpack=False)

    def values(self):
        # TODO this should be an iterator

        data = libmetadata.load_metadata(self.array, unpack=True)
        return data.values()

    def pop(self, key, default=None):
        raise NotImplementedError("dict.pop requires read-write access to array")

    def items(self):
        # TODO this should be an iterator
        data = libmetadata.load_metadata(self.array, unpack=True)
        return tuple((k, data[k]) for k in data.keys())

    def _set_numpy(self, key, value):
        """
        Test helper: directly set meta key-value from a NumPy array.
        Key type and array dimensionality are checked, but no other type-checking
        is done.

        :param key: key
        :param arr: 1d NumPy ndarray
        :return:
        """

        libmetadata._set_metadata_numpy(self.array, key, value)



#%%

import tiledb
from tiledb.libtiledb import ustring
import numpy as np

import unittest, os, time
from tiledb.tests.common import *


class MetadataTest(DiskTestCase):
    def test_metadata_basic(self):
        path = self.path("test_md_basic")

        with tiledb.from_numpy(path, np.ones((5,), np.float64)) as A:
            pass

        # sanity checks
        A = tiledb.open(path)
        A.close()

        # can't read from a closed array
        with self.assertRaises(tiledb.TileDBError):
            A.meta["x"]

        with tiledb.Array(path) as A:
            # can't write to a mode='r' array
            with self.assertRaises(tiledb.TileDBError):
                A.meta["invalid_write"] = 1

            # missing key raises KeyError
            with self.assertRaises(KeyError):
                A.meta["xyz123nokey"]

        # test invalid input
        with tiledb.Array(path, "w") as A:
            # keys must be strings
            with self.assertRaises(ValueError):
                A.meta[123] = 1

            # can't write an int > typemax(Int64)
            with self.assertRaises(OverflowError):
                A.meta["bigint"] = np.iinfo(np.int64).max + 1

            # can't write mixed-type list
            with self.assertRaises(TypeError):
                A.meta["mixed_list"] = [1, 2.1]

            # can't write mixed-type tuple
            with self.assertRaises(TypeError):
                A.meta["mixed_list"] = (0, 3.1)

            # can't write objects
            with self.assertRaises(ValueError):
                A.meta["object"] = object

        test_vals = {
            "int": 10,
            "double": 1.000001212,
            "bytes": b"0123456789abcdeF0123456789abcdeF",
            "str": "abcdefghijklmnopqrstuvwxyz",
            "emptystr": "",
            "tuple_int": (1, 2, 3, 2, 1, int(np.random.randint(0, 10000, 1)[0])),
            "list_int": [1, 2, 3, 2, 1, int(np.random.randint(0, 10000, 1)[0])],
            "tuple_float": (10.0, 11.0, float(np.random.rand(1)[0])),
            "list_float": [10.0, 11.0, float(np.random.rand(1)[0])],
        }

        def tupleize(v):
            if isinstance(v, list):
                v = tuple(v)
            return v

        with tiledb.Array(path, mode="w") as A:
            for k, v in test_vals.items():
                A.meta[k] = v

        with tiledb.Array(path) as A:
            for k, v in test_vals.items():
                # metadata only has one iterable type: tuple, so we cannot
                # perfectly round-trip the input type.

                self.assertEqual(A.meta[k], tupleize(v))

        # test dict-like functionality
        with tiledb.Array(path) as A:
            self.assertSetEqual(set(A.meta.keys()), set(test_vals.keys()))
            self.assertFalse("gnokey" in A.meta)
            self.assertEqual(len(A.meta), len(test_vals))

            for k, v in A.meta.items():
                self.assertTrue(k in test_vals.keys())
                self.assertEqual(tupleize(v), tupleize(test_vals[k]))

        # test a 1 MB blob
        blob = np.random.rand(int((1024 ** 2) / 8)).tostring()
        with tiledb.Array(path, "w") as A:
            A.meta["bigblob"] = blob

        with tiledb.Array(path) as A:
            self.assertEqual(A.meta["bigblob"], blob)
            self.assertEqual(len(A.meta), len(test_vals) + 1)

        # test del key
        with tiledb.Array(path, "w") as A:
            del A.meta["bigblob"]

        with tiledb.Array(path) as A:
            self.assertTrue("bigblob" not in A.meta)
            self.assertEqual(len(A.meta), len(test_vals))
            with self.assertRaises(KeyError):
                A.meta["bigblob"]

        # test pop NotImplementedError
        with tiledb.Array(path, "w") as A:
            with self.assertRaises(NotImplementedError):
                A.meta.pop("nokey", "hello!")

        # Note: this requires a work-around to check all keys
        # test empty value
        with tiledb.Array(path, "w") as A:
            A.meta["empty_val"] = ()

        with tiledb.Array(path) as A:
            self.assertTrue("empty_val" in A.meta)
            self.assertEqual(A.meta["empty_val"], ())

    def test_metadata_consecutive(self):
        ctx = tiledb.Ctx(
            {"sm.vacuum.mode": "array_meta", "sm.consolidation.mode": "array_meta"}
        )
        vfs = tiledb.VFS(ctx=ctx)
        path = self.path("test_md_consecutive")

        write_count = 100

        with tiledb.from_numpy(path, np.ones((5,), np.float64), ctx=ctx) as A:
            pass

        randints = np.random.randint(
            0, int(np.iinfo(np.int64).max) - 1, size=write_count, dtype=np.int64
        )
        randutf8s = [rand_utf8(i) for i in np.random.randint(1, 30, size=write_count)]

        # write 100 times, then consolidate
        for i in range(write_count):
            with tiledb.Array(path, mode="w", ctx=ctx) as A:
                A.meta["randint"] = int(randints[i])
                A.meta["randutf8"] = randutf8s[i]
                time.sleep(0.001)

        with tiledb.Array(path) as A:
            self.assertEqual(A.meta["randint"], randints[-1])
            self.assertEqual(A.meta["randutf8"], randutf8s[-1])

        with tiledb.Array(path, mode="w", ctx=ctx) as aw:
            aw.meta.consolidate()

        with tiledb.Array(path) as A:
            self.assertEqual(A.meta["randint"], randints[-1])
            self.assertEqual(A.meta["randutf8"], randutf8s[-1])

        # use randutf8s as keys, then consolidate
        for _ in range(2):
            for i in range(write_count):
                with tiledb.Array(path, mode="w") as A:
                    A.meta[randutf8s[i] + u"{}".format(randints[i])] = int(randints[i])
                    A.meta[randutf8s[i]] = randutf8s[i]
                    time.sleep(0.001)

        # test data
        with tiledb.Array(path, ctx=ctx) as A:
            for i in range(write_count):
                key_int = randutf8s[i] + u"{}".format(randints[i])
                self.assertEqual(A.meta[key_int], randints[i])
                self.assertEqual(A.meta[randutf8s[i]], randutf8s[i])

        # test expected number of fragments before consolidating
        self.assertEqual(len(vfs.ls(os.path.join(path, "__meta"))), 302)

        with tiledb.Array(path, mode="w", ctx=ctx) as A:
            A.meta.consolidate()

        # test expected number of fragments before vacuuming
        self.assertEqual(len(vfs.ls(os.path.join(path, "__meta"))), 304)

        tiledb.vacuum(path, ctx=ctx)

        # should only have one fragment+'.ok' after vacuuming
        self.assertEqual(len(vfs.ls(os.path.join(path, "__meta"))), 1)

        # test data again after consolidation
        with tiledb.Array(path, ctx=ctx) as A:
            for i in range(write_count):
                key_int = randutf8s[i] + u"{}".format(randints[i])
                self.assertEqual(A.meta[key_int], randints[i])
                self.assertEqual(A.meta[randutf8s[i]], randutf8s[i])

    def test_metadata_small_dtypes(self):
        path = self.path("test_md_small_dtypes")

        with tiledb.from_numpy(path, np.arange(1)) as A:
            pass

        test_vals = {
            "np.int8": np.array((-1,), dtype=np.int8),
            "np.uint8": np.array((2,), dtype=np.uint8),
            "np.int16": np.array((-3,), dtype=np.int16),
            "np.uint16": np.array((4,), dtype=np.uint16),
            "np.int32": np.array((-5,), dtype=np.int32),
            "np.uint32": np.array((6,), dtype=np.uint32),
            "np.float32": np.array((-7.0,), dtype=np.float32),
        }

        with tiledb.Array(path, "w") as A:
            for k, v in test_vals.items():
                A.meta._set_numpy(k, v)

        # note: the goal here is to test read-back of these datatypes,
        #       which is currently done as int for all types
        with tiledb.Array(path) as A:
            for k, v in test_vals.items():
                self.assertEqual(A.meta[k], int(v))



from __future__ import absolute_import

import sys, os, io, re, platform, unittest, random, warnings

import numpy as np
from numpy.testing import assert_array_equal

import tiledb
from tiledb.tests.common import (
    DiskTestCase,
    assert_subarrays_equal,
    rand_utf8,
    rand_ascii,
    rand_ascii_bytes,
)


class BackwardCompatibilityTests(DiskTestCase):
    def test_compat_tiledb_py_0_5_anon_attr_dense(self):
        # Test that anonymous attributes internally stored as "__attr" are presented as ""
        # Normally, we can't actually write an attribute named "__attr" anymore, so we
        # restore a schema written by a patched libtiledb, and rename the attr file.

        schema_data = b"\x05\x00\x00\x00]\x00\x00\x00\x00\x00\x00\x00q\x00\x00\x00\x00\x00\x00\x00\x04\x01\x00\x00\x00\x00\x00\x00\x00\x00\x12\x00\x00\x00\x00\x00\x01\x00\x01\x00\x00\x00\x01\x05\x00\x00\x00\x01\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00q\x00\x00\x009\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00q\x00\x00\x009\x00\x00\x00x\x01ce\x80\x00\x01u(\x83\x81\x11\x08\x19\x18\x98XA\xc4\x7f `\xc0\x10\x01\xc9\x83p\n\x1b\x88\x84\xb0\x81\x8a\xc1l\x88\x00H\x9c\r\x88\xe3\xe3\x13KJ\x8aP\x94\x01\x00\xa2c\x0bD"

        path = self.path("tiledb_py_0_6_anon_attr")
        ctx = tiledb.default_ctx()
        dom = tiledb.Domain(tiledb.Dim(name="d", domain=(0, 0), tile=1, dtype=np.uint8))
        attrs = (tiledb.Attr(name="_attr_", dtype=np.uint8, ctx=ctx),)

        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=False, ctx=ctx)
        tiledb.DenseArray.create(path, schema, ctx=ctx)

        with tiledb.open(path, "w") as A:
            A[0] = 1

        fragment_name = os.path.split(list(A.last_write_info.keys())[0])[-1]
        fragment_path = os.path.join(path, fragment_name)

        # fix up the array the override schema
        with open(os.path.join(path, "__array_schema.tdb"), "wb") as f:
            f.write(schema_data)
        import shutil

        shutil.move(
            os.path.join(fragment_path, "_attr_.tdb"),
            os.path.join(fragment_path, "__attr.tdb"),
        )
        with tiledb.open(path) as A:
            self.assertEqual(A.schema.attr(0).name, "")
            self.assertEqual(A.schema.attr(0)._internal_name, "__attr")
            self.assertEqual(A[0], 1)
            mres = A.multi_index[0]
            self.assertEqual(mres[""], 1)

            qres = A.query(coords=True).multi_index[0]
            self.assertEqual(qres["d"], 0)

    def test_compat_py_0_5_anon_attr_sparse(self):
        import tarfile, base64
        from io import BytesIO

        # This array was written with TileDB-Py 0.5.9:
        # - using the invocation below, followed by
        """
        tiledb.Array.create("path", tiledb.ArraySchema(
                      domain=tiledb.Domain(*[
                      tiledb.Dim(name='d', domain=(0, 2), tile=2, dtype='uint64'),]),
                      attrs=[tiledb.Attr(name='', dtype='int64'),], sparse=True,))
        with tiledb.open("path", 'w') as A:
            A[[0,1,2]] = np.array([1.0,2.0,5.0])
        """
        # - followed by `tar czf array.tgz -C path`
        # - followed by `base64.encodebytes(open("sp6.tgz", 'rb').read())`
        test_array = b"""H4sIANDnmV8AA+2Xz2vUQBTHJ6mLlnpYBGkRD0EQBGV3ZpLJdBFk9bBnj3pKJpvESrsbmo2otyoI
                         Pe/JSy9ePXnwruJBPPYv0P4VRRDNhAxm07o/dBN6eJ9lMpmXSd6Eb96bt602qhyMMcfYyHqbZT3O
                         xwqDmNyyzfRnWwYmFDOCDFb90hB6MkpEnC7l8TCKk2j413lPt4JgZ8pzJl/KWPo6K6LVdpxBkIgq
                         P4OF9Gck1d+kHPSvBan/TtTfbiW+V5WPf9CfM44MXNWCioD+johj8dwZ9beCgajiO5ilP6V2SX9m
                         cdC/Fs6lTQm+q2yaunopO2pIGrSGPGRnhfl30tbMx1rB9kzrC9d1fbd5//yh++HCEcXvXu7/6qJx
                         7/J3fffuZmP/497qgTYOVo6Ojz+Px9d6zfU3r15o6O322r0q3xgoIuOf2NjsULJppVHHSiOPh6Hn
                         9ZnAWFicsk4YspCEOOAd7jFO56kbFq7/KCXEhv2/Dv5bf8cJY/FoEAyTrI70RXJiD5mhPyEWKelv
                         M0Yh/9eBzP+38/PryjZn/pfz19Fk/le2NP/7rvtNFz1D+/Rlb/WrhvQf6Ip0p1KGum1ed3L+Wsmd
                         skl33fQOA+ngYgEXf9ALkyUreX8r77vodKK8P8x7lj/gtXbabOCMsYT8L5Iknvq3Yeb+z6xS/rew
                         bUL+rwMVpRt5K9pUSmjUuiKgTpYQ//0oiv3RlAwwK/7JifrfMjnUf7VQjP+raLJmULYb79s/jY0D
                         hB6kdpUUdHTz4cWspAAAAAAAAAAAAAAA4IzzG7vsp0oAKAAA"""

        path = self.path("test_tiledb_py_0_5_anon_attr_sparse")
        with tarfile.open(fileobj=BytesIO(base64.b64decode(test_array))) as tf:
            tf.extractall(path)
        with tiledb.open(path) as A:
            assert_array_equal(A[:][""], np.array([1.0, 2.0, 5.0]))



import tiledb
from tiledb import Array, ArraySchema, TileDBError
from tiledb.core import increment_stat, use_stats

import os, numpy as np
import sys, time, weakref, warnings
import json
from collections import OrderedDict


def mr_dense_result_shape(ranges, base_shape=None):
    # assumptions: len(ranges) matches number of dims
    if base_shape is not None:
        assert len(ranges) == len(base_shape), "internal error: mismatched shapes"

    new_shape = list()
    for i, rr in enumerate(ranges):
        if rr != ():
            # modular arithmetic gives misleading overflow warning.
            with np.errstate(over="ignore"):
                m = list(
                    map(
                        lambda y: abs(np.uint64(y[1]) - np.uint64(y[0])) + np.uint64(1),
                        rr,
                    )
                )

            new_shape.append(np.sum(m))
        else:
            if base_shape is None:
                raise ValueError(
                    "Missing required base_shape for whole-dimension slices"
                )
            # empty range covers dimension
            new_shape.append(base_shape[i])

    return tuple(new_shape)


def mr_dense_result_numel(ranges):
    return np.prod(mr_dense_result_shape(ranges))


def sel_to_subranges(dim_sel, nonempty_domain=None):
    subranges = list()
    for idx, range in enumerate(dim_sel):
        if np.isscalar(range):
            subranges.append((range, range))
        elif isinstance(range, slice):
            if range.step is not None:
                raise ValueError("Stepped slice ranges are not supported")
            elif range.start is not None and range.stop is not None:
                # we have both endpoints, use them
                rstart = range.start
                rend = range.stop
            else:
                # we are missing one or both endpoints, maybe use nonempty_domain
                if nonempty_domain is None:
                    raise TileDBError(
                        "Open-ended slicing requires a valid nonempty_domain"
                    )
                rstart = nonempty_domain[0] if range.start is None else range.start
                rend = nonempty_domain[1] if range.stop is None else range.stop

            subranges.append((rstart, rend))
        elif isinstance(range, tuple):
            subranges.extend((range,))
        elif isinstance(range, list):
            for el in range:
                subranges.append((el, el))
        else:
            raise TypeError("Unsupported selection ")
    return tuple(subranges)


try:
    import pyarrow

    _have_pyarrow = True
except ImportError:
    _have_pyarrow = False


class MultiRangeIndexer(object):
    """
    Implements multi-range indexing.
    """

    debug = False

    def __init__(self, array, query=None, use_arrow=None):
        if not issubclass(type(array), tiledb.Array):
            raise ValueError("Internal error: MultiRangeIndexer expected tiledb.Array")
        self.array_ref = weakref.ref(array)
        self.schema = array.schema
        self.query = query

        use_arrow_real = isinstance(self, DataFrameIndexer) and _have_pyarrow
        if use_arrow != None:
            use_arrow_real = use_arrow_real and use_arrow

        self.use_arrow = use_arrow_real
        self._preload_metadata: bool = False

    @property
    def array(self):
        assert (
            self.array_ref() is not None
        ), "Internal error: invariant violation (indexing call w/ dead array_ref)"
        return self.array_ref()

    @classmethod
    def __test_init__(cls, array):
        """
        Internal helper method for testing getitem range calculation.
        :param array:
        :return:
        """
        m = cls.__new__(cls)
        m.array_ref = weakref.ref(array)
        m.schema = array.schema
        m.query = None
        return m

    def getitem_ranges(self, idx):
        dom = self.schema.domain
        ndim = dom.ndim
        ned = self.array.nonempty_domain()

        if isinstance(idx, tuple):
            idx = list(idx)
        else:
            idx = [idx]

        ranges = list()
        for i, sel in enumerate(idx):
            if not isinstance(sel, list):
                sel = [sel]
            # don't try to index nonempty_domain if None
            ned_arg = ned[i] if ned else None
            subranges = sel_to_subranges(sel, ned_arg)

            ranges.append(subranges)

        # extend the list to ndim
        if len(ranges) < ndim:
            ranges.extend([tuple() for _ in range(ndim - len(ranges))])

        rval = tuple(ranges)
        return rval

    def __getitem__(self, idx):
        # implements multi-range / outer / orthogonal indexing
        ranges = self.getitem_ranges(idx)
        schema = self.schema
        dom = self.schema.domain
        attr_names = tuple(
            self.schema.attr(i)._internal_name for i in range(self.schema.nattr)
        )
        if schema.sparse:
            dim_names = tuple(dom.dim(i).name for i in range(dom.ndim))
        else:
            dim_names = tuple()

        # set default order
        # - TILEDB_UNORDERED for sparse
        # - TILEDB_ROW_MAJOR for dense
        if self.schema.sparse:
            order = "U"
        else:
            order = "C"

        # if this indexing operation is part of a query (A.query().df)
        # then we need to respect the settings of the query
        if self.query is not None:
            # if we are called via Query object, then we need to respect Query semantics
            if self.query.attrs is not None:
                attr_names = tuple(self.query.attrs)
            else:
                pass  # query.attrs might be None -> all

            if self.query.dims is False:
                dim_names = tuple()
            elif self.query.dims is not None:
                dim_names = tuple(self.query.dims)
            elif self.query.coords is False:
                dim_names = tuple()

            # set query order
            order = self.query.order

        # convert order to layout
        if order is None or order == "C":
            layout = 0
        elif order == "F":
            layout = 1
        elif order == "G":
            layout = 2
        elif order == "U":
            layout = 3
        else:
            raise ValueError(
                "order must be 'C' (TILEDB_ROW_MAJOR), "
                "'F' (TILEDB_COL_MAJOR), "
                "'U' (TILEDB_UNORDERED),"
                "or 'G' (TILEDB_GLOBAL_ORDER)"
            )

        # initialize the pybind11 query object
        from tiledb.core import PyQuery

        q = PyQuery(
            self.array._ctx_(),
            self.array,
            attr_names,
            dim_names,
            layout,
            self.use_arrow,
        )

        q._preload_metadata = self._preload_metadata
        q.set_ranges(ranges)
        q.submit()

        if self.query is not None and self.query.return_arrow:
            return q._buffers_to_pa_table()

        if isinstance(self, DataFrameIndexer) and self.use_arrow:
            return q

        result_dict = OrderedDict(q.results())

        final_names = dict()
        for name, item in result_dict.items():
            if len(item[1]) > 0:
                arr = q.unpack_buffer(name, item[0], item[1])
            else:
                arr = item[0]
                final_dtype = schema.attr_or_dim_dtype(name)
                if len(arr) < 1 and (
                    np.issubdtype(final_dtype, np.bytes_)
                    or np.issubdtype(final_dtype, np.unicode_)
                ):
                    # special handling to get correctly-typed empty array
                    # (expression below changes itemsize from 0 to 1)
                    arr.dtype = final_dtype.str + "1"
                else:
                    arr.dtype = schema.attr_or_dim_dtype(name)
            if name == "__attr":
                final_names[name] = ""
            result_dict[name] = arr

        for name, replacement in final_names.items():
            result_dict[replacement] = result_dict.pop(name)

        if self.schema.sparse:
            return result_dict
        else:
            result_shape = mr_dense_result_shape(ranges, self.schema.shape)
            for arr in result_dict.values():
                # TODO check/test layout
                arr.shape = result_shape
            return result_dict


class DataFrameIndexer(MultiRangeIndexer):
    """
    Implements `.df[]` indexing to directly return a dataframe
    [] operator uses multi_index semantics.
    """

    def __getitem__(self, idx):
        from .dataframe_ import _tiledb_result_as_dataframe, check_dataframe_deps

        check_dataframe_deps()

        idx_start = time.time()

        # we need to use a Query in order to get coords for a dense array
        if not self.query:
            self.query = tiledb.libtiledb.Query(self.array, coords=True)

        self._preload_metadata = True

        result = super(DataFrameIndexer, self).__getitem__(idx)

        if self.use_arrow:
            import pyarrow as pa

            if use_stats():
                pd_start = time.time()

            if isinstance(result, pa.Table):
                # support the `query(return_arrow=True)` mode and return Table untouched
                df = result
            else:
                df = self.pa_to_pandas(result)

            if use_stats():
                pd_duration = time.time() - pd_start
                increment_stat("py.buffer_conversion_time", pd_duration)
                idx_duration = time.time() - idx_start
                increment_stat("py.__getitem__time", idx_duration)
            return df
        else:
            if not isinstance(result, OrderedDict):
                raise ValueError(
                    "Expected OrderedDict result, got '{}'".format(type(result))
                )

            if use_stats():
                pd_start = time.time()

            df = _tiledb_result_as_dataframe(self.array, result)

            if use_stats():
                pd_duration = time.time() - pd_start
                idx_duration = time.time() - idx_start
                tiledb.core.increment_stat("py.buffer_conversion_time", pd_duration)
                tiledb.core.increment_stat("py.__getitem__time", idx_duration)

            return df

    def pa_to_pandas(self, pyquery):
        if not _have_pyarrow:
            raise TileDBError(
                "Cannot convert to pandas via this path without pyarrow; please disable Arrow results"
            )
        try:
            table = pyquery._buffers_to_pa_table()
        except Exception as exc:
            if MultiRangeIndexer.debug:
                print(
                    "Exception during pa.Table conversion, returning pyquery: '{}'".format(
                        exc
                    )
                )
                return pyquery
            raise
        try:
            res_df = table.to_pandas()
        except Exception as exc:
            if MultiRangeIndexer.debug:
                print(
                    "Exception during Pandas conversion, returning (table,query): '{}'".format(
                        exc
                    )
                )
                return table, pyquery
            raise

        if use_stats():
            pd_idx_start = time.time()

        # see also: write path in dataframe_.py
        index_dims = None
        if "__pandas_index_dims" in self.array.meta:
            index_dims = json.loads(self.array.meta["__pandas_index_dims"])

        # see also: write path in dataframe_.py
        attr_reprs = None
        if "__pandas_attribute_repr" in self.array.meta:
            attr_reprs = json.loads(self.array.meta["__pandas_attribute_repr"])

        indexes = list()
        rename_cols = dict()
        for col_name in res_df.columns.values:
            if index_dims and col_name in index_dims:

                # this is an auto-created column and should be unnamed
                if col_name == "__tiledb_rows":
                    rename_cols["__tiledb_rows"] = None
                    indexes.append(None)
                else:
                    indexes.append(col_name)

        if len(rename_cols) > 0:
            res_df.rename(columns=rename_cols, inplace=True)

        if self.query is not None:
            # if we have a query with index_col set, then override any
            # index information saved with the array.
            if self.query.index_col is not True and self.query.index_col is not None:
                res_df.set_index(self.query.index_col, inplace=True)
            elif self.query.index_col is True and len(indexes) > 0:
                # still need to set indexes here b/c df creates query every time
                res_df.set_index(indexes, inplace=True)
            else:
                # don't convert any column to a dataframe index
                pass
        elif len(indexes) > 0:
            res_df.set_index(indexes, inplace=True)

        # apply type translation from TileDB-Py write path
        if attr_reprs:
            res_df = res_df.astype(attr_reprs)

        if use_stats():
            pd_idx_duration = time.time() - pd_idx_start
            tiledb.core.increment_stat("py.pandas_index_update_time", pd_idx_duration)

        return res_df
